@@ -681,6 +681,15 @@ nvswitch_init_lpwr_regs_lr10
                     tempRegVal);
 }
 
+void
+nvswitch_program_l1_scratch_reg_lr10
+(
+    nvswitch_device *device,
+    NvU32 linkNumber
+)
+{
+    // Not Implemented for LR10
+}
 
 void
 nvswitch_init_buffer_ready_lr10
